.text
.balign 16
.globl gcm_init_clmul_rv64i_zbb_zbc
.type gcm_init_clmul_rv64i_zbb_zbc,@function
# Initialize clmul-based implementation of galois field multiplication routine.
# gcm_init_clmul_rv64i_zbb_zbc(ctx->Htable, ctx->H.u)
gcm_init_clmul_rv64i_zbb_zbc:
    # argument 0 = ctx->Htable (store H here)
    # argument 1 = H.u[] (2x 64-bit words) [H_high64, H_low64]

    # Simply store [H_high64, H_low64] for later
    ld      a2,0(a1)
    sd      a2,0(a0)
    ld      a2,8(a1)
    sd      a2,8(a0)

    ret

.text
.balign 16
.globl gcm_gmult_clmul_rv64i_zbb_zbc
.type gcm_gmult_clmul_rv64i_zbb_zbc,@function
# static void gcm_gmult_clmul_rv64i_zbb_zbc(u64 Xi[2], const u128 Htable[16])
# Computes product of X*H mod f
gcm_gmult_clmul_rv64i_zbb_zbc:

    # Load X and H (H is saved previously in gcm_init_clmul_rv64i_zbb_zbc)
    ld              a2,0(a0)
    ld              a3,8(a0)

    ld              a4,0(a1)
    ld              a5,8(a1)

    li              t6,0xe100000000000000

    # Perform Katratsuba Multiplication to generate a 255-bit intermediate
    # A = [A1:A0]
    # B = [B1:B0]
    # Let:
    # [C1:C0] = A1*B1
    # [D1:D0] = A0*B0
    # [E1:E0] = (A0+A1)*(B0+B1)
    # Then:
    # A*B = [C1:C0+C1+D1+E1:D1+C0+D0+E0:D0]

    .word 1803965971
    .word 182851763
    .word 182859827

    .word 1803998867
    .word 183931699
    .word 183939763

    xor             t4,a3,a2
    xor             t5,a5,a4

    .word 200187443
    .word 200192947

    # 0th term is just C1

    # Construct term 1 in E1 (E1 only appears in dword 1)
    xor             t2,t2,t0
    xor             t2,t2,a6
    xor             t2,t2,a7

    # Term 1 is E1

    # Construct term 2 in E0 (E0 only appears in dword 2)
    xor             t3,t3,t1
    xor             t3,t3,a7
    xor             t3,t3,t0

    # Term 2 is E0

    # final term is just D0

    # X*H is now stored in [C1,E1,E0,D0]

    # Left-justify
    slli            a6,a6,1
    # Or in the high bit of E1
    srli            t4,t2,63
    or              a6,a6,t4

    slli            t2,t2,1
    # Or in the high bit of E0
    srli            t5,t3,63
    or              t2,t2,t5

    slli            t3,t3,1
    # Or in the high bit of D0
    srli            t4,t1,63
    or              t3,t3,t4

    slli            t1,t1,1

    # Barrett Reduction
    # c = [E0, D0]
    # We want the top 128 bits of the result of c*f
    # We'll get this by computing the low-half (most significant 128 bits in
    # the reflected domain) of clmul(c,fs)<<1 first, then
    # xor in c to complete the calculation

    # AA = [AA1:AA0] = [E0,D0] = c
    # BB = [BB1:BB0] = [qp_low,0]
    # [CC1:CC0] = AA1*BB1
    # [DD1:DD0] = AA0*BB0
    # [EE1:EE0] = (AA0+AA1)*(BB0+BB1)
    # Then:
    # AA*BB = [CC1:CC0+CC1+DD1+EE1:DD1+CC0+DD0+EE0:DD0]
    # We only need CC0,DD1,DD0,EE0 to compute the low 128 bits of c * qp_low

    .word 201201331
    #clmul          DD0,AA0,BB0     # BB0 is 0, so DD0 = 0
    #clmulh         DD1,AA0,BB0     # BB0 is 0, so DD1 = 0
    xor             t4,t1,t3
    #xor            TEMP2,BB0,BB1   # TEMP2 = BB1 = qp_low
    .word 201234227

    # Result is [N/A:N/A:DD1+CC0+DD0+EE0:DD0]
    # Simplifying: [CC0+EE0:0]
    xor             t5,a3,a4
    # Shift left by 1 to correct for bit reflection
    slli            t5,t5,1

    # xor into c = [E0,D0]
    # Note that only E0 is affected
    xor             t3,t3,t5

    # Now, q = [E0,D0]

    # The final step is to compute clmul(q,[qp_low:0])<<1
    # The leftmost 128 bits are the reduced result.
    # Once again, we use Karatsuba multiplication, but many of the terms
    # simplify or cancel out.
    # AA = [AA1:AA0] = [E0,D0] = c
    # BB = [BB1:BB0] = [qp_low,0]
    # [CC1:CC0] = AA1*BB1
    # [DD1:DD0] = AA0*BB0
    # [EE1:EE0] = (AA0+AA1)*(BB0+BB1)
    # Then:
    # AA*BB = [CC1:CC0+CC1+DD1+EE1:DD1+CC0+DD0+EE0:DD0]
    # We need CC1,CC0,DD0,DD1,EE1,EE0 to compute the leftmost 128 bits of AA*BB


    .word 201201203
    .word 201209523

    #clmul          DD0,AA0,BB0   # BB0 = 0 so DD0 = 0
    #clmulh         DD1,AA0,BB0   # BB0 = 0 so DD1 = 0

    xor             t4,t1,t3
    #xor            TEMP2,BB0,BB1 # BB0 = 0 to TEMP2 == BB1 == qp_low

    .word 201234355
    .word 201242803

    # Need the DD1+CC0+DD0+EE0 term to shift its leftmost bit into the
    # intermediate result.
    # This is just CC0+EE0, store it in TEMP
    xor             t4,a2,a5

    # Result is [CC1:CC0+CC1+EE1:(a single bit)]<<1
    # Combine into [CC1:CC0]
    xor             a2,a2,a3
    xor             a2,a2,a7

    # Shift 128-bit quantity, xor in [C1,E1] and store
    slli            a3,a3,1
    srli            t5,a2,63
    or              a3,a3,t5
    # xor in C1
    xor             a3,a3,a6
    .word 1803998867

    slli            a2,a2,1
    srli            t4,t4,63
    or              a2,a2,t4
    # xor in E1
    xor             a2,a2,t2
    .word 1803965971
    sd              a3,0(a0)
    sd              a2,8(a0)

    ret
